import time
import os
import multiprocessing
from multiprocessing.managers import SharedMemoryManager
from typing import List, Optional, Sequence

import math
from types import SimpleNamespace
import numpy as np
import pandas as pd
from AnyQt.QtCore import QItemSelectionModel, QItemSelection, QItemSelectionRange, Qt
from AnyQt.QtWidgets import QFormLayout, QWidget, QListView, QLabel, QSizePolicy
from AnyQt.QtGui import QDoubleValidator
from scipy.optimize import curve_fit


import Orange.data
from Orange.data import DiscreteVariable, ContinuousVariable, Domain, Variable
from Orange.data.table import Table
from Orange.widgets.widget import OWWidget, Msg, Output, MultiInput
from Orange.widgets import gui, settings

from Orange.widgets.settings import \
    Setting, ContextSetting, DomainContextHandler
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin
from Orange.widgets.data import owconcatenate
from Orange.widgets.data.oweditdomain import disconnected
from orangewidget.utils.listview import ListViewSearch

class DiscDomainModel(DomainModel):

    def data(self, index, role=Qt.DisplayRole):
        value = super().data(index, role)
        if role == Qt.DisplayRole:
            try:
                values = index.data(Qt.UserRole)
            except TypeError:
                pass  # don't have user role (yet)
            else:
                if type(value) is str:
                    value += f" ({values})"
        return value
    
def _restore_selected_items(model, view, setting, connector):
    selection = QItemSelection()
    sel_model: QItemSelectionModel = view.selectionModel()
    with disconnected(sel_model.selectionChanged,
                      connector):
        valid = []
        model_values = model[:]
        for var in setting:
            index = model_values.index(var)
            model_index = view.model().index(index, 0)
            selection.append(QItemSelectionRange(model_index))
            valid.append(var)
        sel_model.select(selection, QItemSelectionModel.ClearAndSelect)


class Results(SimpleNamespace):
    out = None
    model = None
    errorstate = 0

def sort_domain(domain):
    dom = [domain.metas, domain.attributes, domain.class_vars]
    sorted_dom_lst = []
    for i in dom:
        cvs = [[i, j, j.name] for i, j in enumerate(i)]
        rcvs_idx = []
        rcvs = []
        for j, k in enumerate(cvs):
            try:
                cvs[j][-1] = float(k[-1])
            except ValueError:
                rcvs.append(k[1])
                rcvs_idx.append(j)
        for j in reversed(rcvs_idx):
            cvs.pop(j)
        cvs_arr = np.array(cvs)
        if cvs_arr.shape[0] > 0:
            cvs_arr_sorted = cvs_arr[cvs_arr[:,2].argsort()]
            odom_cv = [i[1] for i in cvs_arr_sorted]
            odom = rcvs + odom_cv
        else:
            odom = rcvs
        sorted_dom_lst.append(tuple(odom))
    out = Domain(sorted_dom_lst[1], class_vars=sorted_dom_lst[2], metas=sorted_dom_lst[0])
    return out

def combine_visimg(data, polangles):
    atts = []
    for k, i in enumerate(data):
        try:
            temp = i.attributes['visible_images']
            for j in temp:
                j.name = str(j.name + f' ({polangles[k]} Degrees)')
                atts.append(j)
        except KeyError:
            pass
        except AttributeError:
            pass
    attsdict = {'visible_images': atts}
    return attsdict

def run(data, feature, alphas, map_x, map_y, invert_angles, polangles, average,
        sep, state: TaskState):

    results = Results()

    output, model, spectra, origmetas, errorstate = process_polar_abs(data, alphas, feature, map_x,
                                                    map_y, invert_angles, polangles, average, state)


    tempoutaddmetas = [[ContinuousVariable.make('Azimuth Angle (' + i.name + ')'),
                    ContinuousVariable.make('Hermans Orientation Function (' + i.name + ')'),
                    ContinuousVariable.make('Intensity (' + i.name + ')'),
                    ContinuousVariable.make('Amplitude (' + i.name + ')'),
                    ContinuousVariable.make('R-squared (' + i.name + ')')] for i in feature]
    outaddmetas = []
    for i in tempoutaddmetas:
        outaddmetas = outaddmetas + i

    tempmodaddmetas = [[ContinuousVariable.make('R-squared (' + i.name + ')'),
                    ContinuousVariable.make('a0 (' + i.name + ')'),
                    ContinuousVariable.make('a1 (' + i.name + ')'),
                    ContinuousVariable.make('a2 (' + i.name + ')')] for i in feature]
    modaddmetas = []
    for i in tempmodaddmetas:
        modaddmetas = modaddmetas + i
    ometadom = data[0].domain.metas
    if average is False:
        values = tuple(f'{i} Degrees' for i in polangles)
        PolAng = DiscreteVariable.make('Polarization Angle', values=values)
        outmetadom = (ometadom + tuple([PolAng]) + tuple(outaddmetas))
        modmetadom = (ometadom + tuple([PolAng]) + tuple(modaddmetas))
        output_stack = tuple(output for i in polangles)
        model_stack = tuple(model for i in polangles)
        output = np.vstack(output_stack)
        model = np.vstack(model_stack)
    elif average is True:
        if sep is not None:
            sep_idx = ometadom.index(sep)
            try:
                origmetas = np.c_[origmetas[:,0:sep_idx], origmetas[:,sep_idx+1:]]
            except ValueError:
                origmetas = np.r_[origmetas[:,0:sep_idx]]
            ometadom = tuple(i for i in ometadom if i is not sep)
        outmetadom = (ometadom + tuple(outaddmetas))
        modmetadom = (ometadom + tuple(modaddmetas))

    if state.is_interruption_requested():
        results = None
        return results

    ofeatdom = data[0].domain.attributes
    datadomain = Domain(ofeatdom, metas = outmetadom)
    moddomain = Domain(ofeatdom, metas = modmetadom)
    outmetas = np.hstack((origmetas, output))
    modmetas = np.hstack((origmetas, model))

    out = Table.from_numpy(datadomain, X=spectra, Y=None, metas=outmetas)
    mod = Table.from_numpy(moddomain, X=spectra, Y=None, metas=modmetas)

    if state.is_interruption_requested():
        results = None
        return results

    results.out = out
    results.model = mod
    results.errorstate = errorstate

    attsdict = combine_visimg(data, polangles)

    results.out.attributes = attsdict
    results.model.attributes = attsdict
    return results

#Calculate by fitting to function
def azimuth(x,a0,a1,a2):
    t = 2*np.radians(x)
    return a0*np.sin(t)+a1*np.cos(t)+a2

def azimuth_jac(x, a0, a1, a2):
    t = 2*np.radians(x).reshape(-1, 1)
    da0 = np.sin(t)
    da1 = np.cos(t)
    da2 = np.ones(t.shape)
    return np.hstack((da0, da1, da2))

def calc_angles(a0,a1):
    return np.degrees(0.5*np.arctan(a0/a1))

def ampl2(a0,a1):
    return (2*(math.sqrt(a0**2+a1**2)))

def orfunc(alpha,a0,a1,a2):
    if alpha < 54.73:
        Dmax = (2*a2+2*math.sqrt(a0**2+a1**2))/(2*a2-2*math.sqrt(a0**2+a1**2))
        return ((Dmax-1)/(Dmax+2)*(2/(3*np.cos(np.radians(alpha))**2-1)))
    elif alpha >= 54.73:
        Dmin = (2*a2-2*math.sqrt(a0**2+a1**2))/(2*a2+2*math.sqrt(a0**2+a1**2))
        return ((Dmin-1)/(Dmin+2)*(2/(3*np.cos(np.radians(alpha))**2-1)))
    return None

def find_az(params):
    Az0 = calc_angles(params[0],params[1])
    Abs0 = azimuth(Az0, *params)
    Az1 = calc_angles(params[0],params[1])+90
    Abs1 = azimuth(Az1, *params)
    Az2 = calc_angles(params[0],params[1])-90


    if Abs0 > Abs1:
        Az = Az0
    elif Abs1 > Abs0:
        if Az1 < 90:
            Az = Az1
        elif Az1 > 90:
            Az = Az2
    return Az

def compute(xys, yidx, smms, shapes, dtypes, polangles):
    tcvs = smms[0]
    cvs = np.ndarray(shapes[0], dtype=dtypes[0], buffer=tcvs.buf)
    tout = smms[3]
    out = np.ndarray(shapes[3], dtype=dtypes[3], buffer=tout.buf)
    tmod = smms[4]
    mod = np.ndarray(shapes[4], dtype=dtypes[4], buffer=tmod.buf)
    tcoords = smms[5]
    coords = np.ndarray(shapes[5], dtype=dtypes[5], buffer=tcoords.buf)
    tvars = smms[6]
    vars = np.ndarray(shapes[6], dtype=dtypes[6], buffer=tvars.buf)

    x = np.asarray(polangles)

    for i in range(yidx[0], yidx[1]):#y-values(rows)
        if vars[-1] != 0:
            break
        for j in enumerate(xys[0]):#x-values(cols)
            if vars[-1] != 0:
                break
            for l in range(cvs.shape[2]):
                if np.any(np.isnan(cvs[i,j[0],l,:]), axis=0):
                    continue
                out[i,j[0],l,0] = coords[i,j[0],1]#x-map
                mod[i,j[0],l,0] = coords[i,j[0],1]
                out[i,j[0],l,1] = coords[i,j[0],0]#y-map
                mod[i,j[0],l,1] = coords[i,j[0],0]

                temp = list(cvs[i,j[0],l,:])

                params = curve_fit(azimuth, x, temp, jac=azimuth_jac)[0]

                residuals = temp - azimuth(x, *params)
                ss_res = np.sum(residuals**2)
                ss_tot = np.sum((temp-np.mean(temp))**2)
                if ss_tot == 0:
                    vars[-1] = 1
                    break
                out[i,j[0],l,6] = 1-(ss_res/ss_tot)
                mod[i,j[0],l,2] = 1-(ss_res/ss_tot)
                out[i,j[0],l,2] = find_az(params)
                out[i,j[0],l,3] = orfunc(vars[l], *params)
                out[i,j[0],l,4] = params[2]
                out[i,j[0],l,5] = ampl2(params[0],params[1])
                mod[i,j[0],l,3] = params[0]
                mod[i,j[0],l,4] = params[1]
                mod[i,j[0],l,5] = params[2]


def unique_xys(images, map_x, map_y):
    lsxs = np.empty(0)
    lsys = np.empty(0)
    for i in enumerate(images):
        tempdata = i[1].transform(Domain([map_x, map_y]))
        lsx = np.unique(tempdata.X[:,0])
        lsy = np.unique(tempdata.X[:,1])
        lsxs = np.append(lsxs, lsx)
        lsys = np.append(lsys, lsy)

    ulsxs = np.unique(lsxs)
    ulsys = np.unique(lsys)
    return ulsxs, ulsys

def start_compute(ulsxs, ulsys, names, shapes, dtypes, polangles, state):
    # single core processing is faster for small data sets and small number of selected features
    # if <data size> > x:
    # ncpu = os.cpu_count()
    # ncpu = 6
    env_proc = os.getenv('QUASAR_N_PROCESSES')
    ncpu = os.cpu_count() if env_proc == "all" else int(env_proc) if env_proc else min(2, os.cpu_count())
    tulsys = np.array_split(ulsys, ncpu)
    state.set_status("Calculating...")
    threads=[]
    cumu = 0
    for i in range(ncpu):
        tlsxys = [ulsxs,tulsys[i]]
        yidx = [cumu, cumu+len(tulsys[i])]
        cumu += len(tulsys[i])
        # compute(tlsxys, yidx, shapes, dtypes, polangles, i)
        t = multiprocessing.Process(target=compute,
                                    args=(tlsxys, yidx, names, shapes, dtypes, polangles))
        threads.append(t)
        t.start()

    # for t in threads:
    #     t.join()

    # else:
        # ncpu = 1
        # tulsys = np.array_split(ulsys, ncpu)
        # state.set_status("Calculating...")
        # threads=[]
        # cumu = 0
        # for i in range(ncpu):
        #     tlsxys = [ulsxs,tulsys[i]]
        #     yidx = [cumu, cumu+len(tulsys[i])]
        #     cumu += len(tulsys[i])
        #     compute(tlsxys, yidx, shapes, dtypes, polangles, i)
    return threads

def process_polar_abs(images, alphas, feature, map_x, map_y, invert, polangles, average, state):
    state.set_status("Preparing...")

    ulsxs, ulsys = unique_xys(images, map_x, map_y)

    if len(ulsxs) > 1:
        dx = np.sum(np.diff(ulsxs))/(len(ulsxs)-1)
    else:
        dx = 1
    if len(ulsys) > 1:
        dy = np.sum(np.diff(ulsys))/(len(ulsys)-1)
    else:
        dy = 1
    minx = np.min(ulsxs)
    miny = np.min(ulsys)

    featnames = [i.name for i in feature]
    cvs = np.full((np.shape(ulsys)[0], np.shape(ulsxs)[0], len(featnames), len(images)), np.nan)
    spec = np.full((np.shape(ulsys)[0], np.shape(ulsxs)[0],
                    images[0].X.shape[1], len(images)), np.nan, dtype=object)
    metas = np.full((np.shape(ulsys)[0], np.shape(ulsxs)[0],
                     images[0].metas.shape[1], len(images)), np.nan, dtype=object)
    out = np.full((np.shape(ulsys)[0], np.shape(ulsxs)[0], len(featnames), 7), np.nan)
    mod = np.full((np.shape(ulsys)[0], np.shape(ulsxs)[0], len(featnames), 6), np.nan)
    coords = np.full((np.shape(ulsys)[0], np.shape(ulsxs)[0], 2), np.nan)
    vars = np.hstack((np.asarray(alphas),0))
    fill = np.full((np.shape(ulsys)[0], np.shape(ulsxs)[0]), np.nan)

    for i, j in enumerate(images):
        if state.is_interruption_requested():
            return None, None, None, None, 2
        cv = [j.domain[k] for k in featnames]
        doms = [map_x, map_y] + cv
        tempdata: Table = j.transform(Domain(doms))
        temp_xy = tempdata.X[:,0:2].copy()
        temp_xy[:,0] = np.rint(((temp_xy[:,0]-minx)/dx))
        temp_xy[:,1] = np.rint(((temp_xy[:,1]-miny)/dy))
        temp_xy = np.array(temp_xy, dtype=np.int_)
        cvs[temp_xy[:,1],temp_xy[:,0],:,i] = tempdata[:,2:]
        spec[temp_xy[:,1],temp_xy[:,0],:,i] = j.X
        metas[temp_xy[:,1],temp_xy[:,0],:,i] = j.metas
    xys = pd.DataFrame(fill, index=ulsys, columns=ulsxs, dtype=object)
    for k, i in enumerate(xys.index):
        for l, j in enumerate(xys.columns):
            coords[k,l,0] = i
            coords[k,l,1] = j

    if state.is_interruption_requested():
        return None, None, None, None, 2

    with SharedMemoryManager() as smm:
        tcvs = smm.SharedMemory(size=cvs.nbytes)
        scvs = np.ndarray(cvs.shape, dtype=cvs.dtype, buffer=tcvs.buf)
        scvs[:,:,:] = cvs[:,:,:]
        tout = smm.SharedMemory(size=out.nbytes)
        sout = np.ndarray(out.shape, dtype=out.dtype, buffer=tout.buf)
        sout[:,:,:,:] = out[:,:,:,:]
        tmod = smm.SharedMemory(size=mod.nbytes)
        smod = np.ndarray(mod.shape, dtype=mod.dtype, buffer=tmod.buf)
        smod[:,:,:,:] = mod[:,:,:,:]
        tcoords = smm.SharedMemory(size=coords.nbytes)
        scoords = np.ndarray(coords.shape, dtype=coords.dtype, buffer=tcoords.buf)
        scoords[:,:,:] = coords[:,:,:]
        tvars = smm.SharedMemory(size=vars.nbytes)
        svars = np.ndarray(vars.shape, dtype=vars.dtype, buffer=tvars.buf)
        svars[:] = vars[:]

        smms = [tcvs, None, None, tout, tmod, tcoords, tvars]
        shapes = [cvs.shape, spec.shape, metas.shape, out.shape, mod.shape, coords.shape, vars.shape]
        dtypes = [cvs.dtype, spec.dtype, metas.dtype, out.dtype, mod.dtype, coords.dtype, vars.dtype]

        if state.is_interruption_requested():
            return None, None, None, None, 2
        threads = start_compute(ulsxs, ulsys, smms, shapes, dtypes, polangles, state)

        for t in threads:
            t.join(0)

        while any([i.exitcode == None for i in threads]):
            if state.is_interruption_requested():
                svars[1] = 2
            time.sleep(0.10)

        state.set_status("Finishing...")
        if invert is True:
            sout[:,:,:,2] = sout[:,:,:,2]*-1
        outputs = np.reshape(sout[:,:,:,2:], (np.shape(ulsys)[0]*np.shape(ulsxs)[0], 5*len(featnames)))
        model = np.reshape(smod[:,:,:,2:], (np.shape(ulsys)[0]*np.shape(ulsxs)[0], 4*len(featnames)))
        vars[:] = svars[:]

    if state.is_interruption_requested():
        return None, None, None, None, 2

    spectra = []
    met = []
    if average is False:
        for i in range(len(polangles)):
            spectratemp = np.reshape(spec[:,:,:,i],
                                    (np.shape(ulsys)[0]*np.shape(ulsxs)[0], images[0].X.shape[1]))
            spectratemp = spectratemp[~np.isnan(model).any(axis=1)]
            spectra.append(spectratemp)
            metatemp = np.reshape(metas[:,:,:,i],
                                (np.shape(ulsys)[0]*np.shape(ulsxs)[0], images[0].metas.shape[1]))
            metatemp = metatemp[~np.isnan(model).any(axis=1)]
            metatemp = np.append(metatemp, np.full((np.shape(metatemp)[0],1), i), axis=1)
            met.append(metatemp)

            if state.is_interruption_requested():
                return None, None, None, None, 2

    elif average is True:
        average_spec = np.average(spec, axis=3)
        spectratemp = np.reshape(average_spec,
                                (np.shape(ulsys)[0]*np.shape(ulsxs)[0], images[0].X.shape[1]))
        spectratemp = spectratemp[~np.isnan(model).any(axis=1)]
        spectra.append(spectratemp)
        metatemp = np.reshape(metas[:,:,:,0],
                            (np.shape(ulsys)[0]*np.shape(ulsxs)[0], images[0].metas.shape[1]))
        metatemp = metatemp[~np.isnan(model).any(axis=1)]
        met.append(metatemp)

    outputs = outputs[~np.isnan(model).any(axis=1)]
    model = model[~np.isnan(model).any(axis=1)]

    spectra = np.concatenate((spectra), axis=0)
    meta = np.concatenate((met), axis=0)

    return outputs, model, spectra, meta, vars[-1]


class OWPolar(OWWidget, ConcurrentWidgetMixin):

    # Widget's name as displayed in the canvas
    name = "4+ Angle Polarization"

    # Short widget description
    description = (
        "Calculate Azimuth Angle, Orientation function, Amplitude and Intensity of "
        "vibrational mode(s) using polarized data measured at 4 or more polarization angles.")

    icon = "icons/polar.svg"

    settings_version = 2

    # Define inputs and outputs
    class Inputs:
        data = MultiInput("Data", Orange.data.Table, default=True)

    class Outputs:
        polar = Output("Polar Data", Orange.data.Table, default=True)
        model = Output("Curve Fit model data", Orange.data.Table)

    autocommit = settings.Setting(False)

    settingsHandler = DomainContextHandler()

    want_main_area = False
    resizing_enabled = True
    alpha = Setting(0, schema_only=True)

    map_x = ContextSetting(None, exclude_attributes=True, exclude_class_vars=True)
    map_y = ContextSetting(None, exclude_attributes=True, exclude_class_vars=True)
    invert_angles = Setting(False, schema_only=True)
    average = Setting(False, schema_only=True)
    angles = ContextSetting(None, exclude_attributes=True, exclude_class_vars=True)
    spec_type = Setting(0)
    alphas: List[float] = ContextSetting([])
    feats: List[Variable] = ContextSetting([])

    class Warning(OWWidget.Warning):
        # nodata = Msg("No useful data on input!")
        # noang = Msg("Must receive 4 angles at specified polarisation")
        nofeat = Msg("Select one or more Features to calculate orientations with")
        noxy = Msg("Select X and Y variables")
        pol = Msg("Invalid Polarization angles")
        notenough = Msg("Must have >= 4 angles")
        wrongdata = Msg("Model returns inf. Inappropriate data")
        # tomany = Msg("Widget must receive data at data input or discrete angles only")
        # missingfeat = Msg("All inputs must have the selected feature")
        # renamed_variables = Msg("Variables with duplicated names have been renamed.")
        XYfeat = Msg("Selected feature(s) cannot be the same as XY selection")

    class Information(OWWidget.Information):
        meta_calc = Msg("Meta and Target variables are not transformed to absorptance during calculations")

    def __init__(self):
        super().__init__()
        ConcurrentWidgetMixin.__init__(self)
        gui.OWComponent.__init__(self)

        self.anglst = []
        self.lines = []
        self.labels = []
        self.multiin_anglst = []
        self.multiin_lines = []
        self.multiin_labels = []
        self.polangles = []
        self.n_inputs = 0

        self._dumb_tables = owconcatenate.OWConcatenate._dumb_tables
        self._get_part = owconcatenate.OWConcatenate._get_part
        self.merge_domains = owconcatenate.OWConcatenate.merge_domains

        self._data_inputs: List[Optional[Table]] = []

        hbox = gui.hBox(self.controlArea)
        #col 1

        vbox2 = gui.vBox(hbox, None)

        form2 = QWidget()
        formlayout2 = QFormLayout()
        form2.setLayout(formlayout2)

        self.multifile = gui.widgetBox(vbox2, "Single input (with all angles)",
                                       sizePolicy=(QSizePolicy.Minimum, QSizePolicy.Fixed))

        self.anglemetas = DomainModel(DomainModel.METAS, valid_types=DiscreteVariable)
        self.anglesel = gui.comboBox(self.multifile, self, 'angles', searchable=True,
                                     label='Select Angles by:', callback=self._change_angles,
                                     model=self.anglemetas)
        self.anglesel.setDisabled(True)


        self.multiin = gui.widgetBox(vbox2, "Multiple inputs (single angle per input)",
                                     sizePolicy=(QSizePolicy.Minimum, QSizePolicy.Fixed))

        vbox2.layout().addWidget(form2)

        #col 2
        vbox1 = gui.vBox(hbox, "Features")

        self.featureselect = DiscDomainModel(DomainModel.SEPARATED,
            valid_types=ContinuousVariable)
        self.feat_view = ListViewSearch(selectionMode=QListView.ExtendedSelection)
        self.feat_view.setModel(self.featureselect)
        self.feat_view.selectionModel().selectionChanged.connect(self._feat_changed)
        vbox1.layout().addWidget(self.feat_view)
        vbox1.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum))
        self.contextOpened.connect(
            lambda: _restore_selected_items(model=self.featureselect,
                                            view=self.feat_view,
                                            setting=self.feats,
                                            connector=self._feat_changed))

        gui.button(vbox1, self, "Don't use selected features",
                   callback=self.remove_feat)

        #col 3
        vbox = gui.vBox(hbox, None)

        form = QWidget()
        formlayout = QFormLayout()
        form.setLayout(formlayout)

        xybox = gui.widgetBox(vbox, "Data XY Selection",
                              sizePolicy=(QSizePolicy.Minimum, QSizePolicy.Fixed))

        self.x_axis = DomainModel(DomainModel.METAS, valid_types=ContinuousVariable)
        self.y_axis = DomainModel(DomainModel.METAS, valid_types=ContinuousVariable)

        gui.comboBox(xybox, self, 'map_x', searchable=True, label="X Axis",
            callback=self._change_input, model=self.x_axis)
        gui.comboBox(xybox, self, 'map_y', searchable=True, label="Y Axis",
            callback=self._change_input, model=self.y_axis)

        vbox.layout().addWidget(form)

        specbox = gui.widgetBox(vbox, "Spectra Type",
                                sizePolicy=(QSizePolicy.Minimum, QSizePolicy.Fixed))

        self.spec_b1 = gui.radioButtons(specbox, self, 'spec_type', ['Absorptance', 'Absorbance', 'Transmittance'],
                                        callback=self._process_spectype)

        vbox.layout().addWidget(form)

        pbox = gui.widgetBox(vbox, "Parameters", sizePolicy=(QSizePolicy.Minimum, QSizePolicy.Fixed))
        self.alphaedit = gui.lineEdit(pbox, self, "alpha", u"TDM Tilt (\N{DEGREE SIGN})",
                     callback=self.change_alphas, valueType=float,
                     validator=QDoubleValidator(0.00, 90.00, 2), tooltip= \
                         "The angle (in degrees) between the long axis of the molecule and the transition dipole moment")

        gui.checkBox(pbox, self, 'invert_angles', label="Invert Angles",
                     callback=self._change_input)

        gui.checkBox(pbox, self, 'average', label='Average Spectra',
                     callback=self._change_input)

        gui.auto_commit(self.controlArea, self, "autocommit", "Apply", commit=self.commit)

        gui.rubber(vbox2)
        cbox = gui.widgetBox(vbox2, "Citations")
        citation = gui.widgetLabel(cbox)
        citation.setOpenExternalLinks(True)
        citation.setText('\
            <body>\
            When publishing results, consider citing:<br> \
            - <a href="https://doi.org/10.1021/ma302560q">Hikima et al. (2013)</a><br> \
            - <a href="https://doi.org/10.1016/j.compbiomed.2025.110573">Gassner et al. (2025)</a> \
            </body>')

        self._change_input()
        self.contextAboutToBeOpened.connect(lambda x: self.init_attr_values(x[0]))

    def sort_row(self, unsorted):
        row, feats, alphas = list(zip(*[unsorted]))
        return row

    def sort_feats(self):
        model = self.feat_view.model()
        rows = [model.indexOf(row) for row in self.feats]
        featalphas = list(zip(rows, self.feats, self.alphas))
        temp = sorted(featalphas, key=self.sort_row)
        rows, feats, alphas = list(zip(*temp))
        self.feats = list(feats)
        self.alphas = list(alphas)

    def _feat_changed(self):
        self.Warning.nofeat.clear()
        rows = self.feat_view.selectionModel().selectedRows()
        model = self.feat_view.model()
        for row in rows:
            if model[:][row.row()] not in self.feats:
                self.feats.append(model[:][row.row()])
                self.alphas.append(self.alpha)
                model.setData(model.index(row.row()),
                              f'TDM = {self.alpha}\N{DEGREE SIGN}',
                              role=Qt.UserRole)
        self.sort_feats()
        if len(rows) > 0:
            self.Warning.nofeat.clear()
        else:
            self.Warning.nofeat()
            return
        self.commit.deferred()

    def remove_feat(self):
        rows = self.feat_view.selectionModel().selectedRows()
        model = self.feat_view.model()
        for row in rows:
            idx = self.feats.index(model[:][row.row()])
            self.feats.remove(model[:][row.row()])
            self.alphas.pop(idx)
            model.setData(model.index(row.row()), 'Not used', role=Qt.UserRole)
        self.commit.deferred()

    def change_alphas(self):
        model = self.feat_view.model()
        rows = self.feat_view.selectionModel().selectedRows()
        for row in rows:
            try:
                idx = self.feats.index(model[:][row.row()])
            except ValueError:
                self._feat_changed()
                idx = self.feats.index(model[:][row.row()])
            self.alphas[idx] = self.alpha
            model.setData(model.index(row.row()), f"TDM = {self.alpha}\N{DEGREE SIGN}", role=Qt.UserRole)
        self.commit.deferred()

    def restore_alphas(self):
        rows = self.feat_view.selectionModel().selectedRows()
        model = self.feat_view.model()
        for row in rows:
            idx = self.feats.index(model[:][row.row()])
            model.setData(model.index(row.row()), f"TDM = {self.alphas[idx]}\N{DEGREE SIGN}", role=Qt.UserRole)

    def init_attr_values(self, data):
        domain = data.domain if data is not None else None
        self.featureselect.set_domain(domain)
        self.x_axis.set_domain(domain)
        self.y_axis.set_domain(domain)
        self.anglemetas.set_domain(domain)
        self.angles = self.anglemetas[0] if self.anglemetas else None
        self.map_x = self.x_axis[0] if self.x_axis else None
        self.map_y = self.y_axis[1] if len(self.y_axis) >= 2 \
            else self.map_x
        self.restore_alphas()

    def _change_input(self):
        self.commit.deferred()

    def _change_angles(self):
        # self.Warning.nodata.clear()
        if self.angles:
            self.clear_angles(self.anglst, self.lines, self.labels, self.multifile)
            self.anglst = []
            self.lines = []
            self.labels = []
            self.Warning.notenough.clear()
            if len(self.angles.values) < 4:
                self.Warning.notenough()
            else:
                tempangles = np.linspace(0, 180, len(self.angles.values)+1)
                for i, j in enumerate(self.angles.values):
                    self.add_angles(self.anglst, j, self.labels, self.lines, self.multifile,
                                    i, tempangles[i], self._send_angles)
                self._send_angles()
                for i in self.labels:
                    i.setDisabled(False)
                for i in self.lines:
                    i.setDisabled(False)
                self.commit.deferred()

    def add_angles(self, anglst, lab, labels, lines, widget,
                   i, place, callback): #to be used in a loop
        file = os.path.basename(lab)
        anglst.append(lab)
        ledit = gui.lineEdit(widget, self, None, label = file, callback = callback)
        ledit.setText(str(place))
        lines.append(ledit)
        for j in ledit.parent().children():
            if isinstance(j, QLabel):
                labels.append(j)

    def clear_angles(self, anglst, lines, labels, widget):
        if widget is self.multiin:
            for i in reversed(range(self.multiin.layout().count())):
                self.multiin.layout().itemAt(i).widget().setParent(None)
        if widget is self.multifile:
            for i in reversed(range(self.multifile.layout().count())):
                if i != 0:
                    self.multifile.layout().itemAt(i).widget().setParent(None)
        anglst.clear()
        lines.clear()
        labels.clear()
        self.polangles.clear()

    def _send_ind_angles(self):
        self.polangles.clear()
        for i in self.multiin_lines:
            self.polangles.append(i.text())
        try:
            pol = []
            for i in self.polangles:
                pol.append(float(i))
            self.polangles = pol
            self.commit.deferred()
        except ValueError:
            pass

    def _send_angles(self):
        self.polangles.clear()
        for i in self.lines:
            self.polangles.append(i.text())
        try:
            pol = []
            for i in self.polangles:
                pol.append(float(i))
            self.polangles = pol
            self.commit.deferred()
        except ValueError:
            pass

    def input_select(self):
        self.angles = None
        if len(self.data) == 0 or 1 < len(self.data) < 4:
            self.anglesel.setDisabled(True)
            for i in self.multiin_labels:
                i.setDisabled(True)
            for i in self.multiin_lines:
                i.setDisabled(True)
        elif len(self.data) == 1:
            self.anglesel.setDisabled(False)
            for i in self.multiin_labels:
                i.setDisabled(True)
            for i in self.multiin_lines:
                i.setDisabled(True)
        elif len(self.data) > 3:
            self.anglesel.setDisabled(True)
            for i in self.multiin_labels:
                i.setDisabled(False)
            for i in self.multiin_lines:
                i.setDisabled(False)
            self._send_ind_angles()

    def _process_spectype(self):
        self.commit.deferred()

    def check_params(self):
        self.Warning.nofeat.clear()
        if self.feats is None or len(self.feats) == 0:
            self.Warning.nofeat()
            return
        self.Warning.noxy.clear()
        if self.map_x is None or self.map_y is None:
            self.Warning.noxy()
            return
        self.Warning.pol.clear()
        if len(self.polangles) == 0:
            self.Warning.pol()
            return
        for i in self.polangles:
            if isinstance(i, float) is False:
                self.Warning.pol()
                return
        self.Warning.XYfeat.clear()
        for i in self.feats:
            if i in (self.map_x, self.map_y):
                self.Warning.XYfeat()
                return
        self.Warning.wrongdata.clear()

    @Inputs.data
    def set_data(self, index: int, dataset: Table):
        self._data_inputs[index] = dataset

    @Inputs.data.insert
    def insert_data(self, index, dataset):
        self._data_inputs.insert(index, dataset)
        self.n_inputs += 1
        self.idx = index

    @Inputs.data.remove
    def remove_data(self, index):
        self._data_inputs.pop(index)
        self.n_inputs -= 1
        self.polangles.clear()

    @property
    def more_data(self) -> Sequence[Table]:
        return [t for t in self._data_inputs if t is not None]

    def handleNewSignals(self):
        self.cancel()
        self.check_and_close_context()
        self.data = None
        self.Warning.clear()
        self.Outputs.polar.send(None)
        self.Outputs.model.send(None)
        self.data = self.more_data

        self.clear_angles(self.anglst, self.lines, self.labels, self.multifile)
        self.clear_angles(self.multiin_anglst, self.multiin_lines,
                          self.multiin_labels, self.multiin)

        names = [i.name for i in self.data]

        tempangles = np.linspace(0, 180, len(self.data)+1)
        for i in range(len(self.data)):
            self.add_angles(self.multiin_anglst, names[i], self.multiin_labels,
                            self.multiin_lines, self.multiin, i, tempangles[i],
                            self._send_ind_angles)

        self.input_select()

        if len(self.data) == 0:
            self.Outputs.polar.send(None)
            self.Outputs.model.send(None)
            self.contextAboutToBeOpened.emit([Table.from_domain(Domain(()))])
            return

        if len(self.data) == 1:
            self.openContext(self.data[0])
            self.angles = self.anglemetas[0] if self.anglemetas else None #Fixes self.angles being None if it was first a text variable and then changed to categorical, but I feel this could be done better...
            self._change_angles()
        elif 1 < len(self.data) < 4 or len(self.data) == 0:
            self.Warning.notenough()
            self.contextAboutToBeOpened.emit([Table.from_domain(Domain(()))])
            return
        else:
            tables = self._dumb_tables(self)
            domains = [table.domain for table in tables]
            self._get_part = self._get_part
            self.merge_type = 0
            domain1 = self.merge_domains(self, domains)
            domain1 = sort_domain(domain1)
            self.merge_type = 1
            domain2 = self.merge_domains(self, domains)

            self.sorted_data = [table.transform(domain1) for table in tables]
            self.openContext(Table.from_domain(domain2))

        for i in range(len(self.feat_view.model())):
            self.feat_view.model().setData(self.feat_view.model().index(i), 'Not used', role=Qt.UserRole)
        self.restore_alphas()

        self.commit.now()

    @gui.deferred
    def commit(self):
        self.cancel()
        self.check_params()
        self.Information.meta_calc.clear()
        if len(self.Warning.active) > 0:
            return

        if len(self.data) == 1:
            if self.angles:
                fncol = self.data[0][:, self.angles.name].metas.reshape(-1)
                images = []
                for fn in self.anglst:
                    images.append(self.data[0][self.angles.to_val(fn) == fncol])
                sorted_data = images
            else:
                return
        elif 1 < len(self.data) < 4:
            self.Warning.notenough()
            self.Outputs.polar.send(None)
            self.Outputs.model.send(None)
            return
        else:
            sorted_data = self.sorted_data

        if self.spec_type != 0:
            if self.spec_type == 1:
                for i, j in enumerate(sorted_data):
                    sorted_data[i].X = 1-np.power(10, j.X*-1)
            elif self.spec_type == 2:
                for i, j in enumerate(sorted_data):
                    sorted_data[i].X = 1-j.X
            if any(i not in sorted_data[0].domain.attributes for i in self.feats):
                self.Information.meta_calc()

        self.start(run, sorted_data, list(self.feats), self.alphas, self.map_x,
                   self.map_y, self.invert_angles, list(self.polangles),
                   self.average, self.angles)

    def on_done(self, result: Results):
        if result is None:
            self.Outputs.polar.send(None)
            self.Outputs.model.send(None)
            return
        if result.errorstate == 1:
            self.Warning.wrongdata()
        else:
            if self.spec_type != 0:
                if self.spec_type == 1:
                    result.out.X = -np.log10(1-result.out.X)
                    result.model.X = -np.log10(1-result.model.X)
                elif self.spec_type == 2:
                    result.out.X = 1-result.out.X
                    result.model.X = 1-result.model.X
            self.Outputs.polar.send(result.out)
            self.Outputs.model.send(result.model)

    def on_partial_result(self, result):
        pass

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()

    def check_and_close_context(self):
        if hasattr(self, 'data'):
            if len(self.data) == 1:
                self.closeContext()
            else:
                self.angles = None
                self.closeContext()


if __name__ == "__main__":  # pragma: no cover
    from Orange.widgets.utils.widgetpreview import WidgetPreview
    import orangecontrib.spectroscopy  # so that can be loaded
    WidgetPreview(OWPolar).run(insert_data=[(0, Orange.data.Table("polar/4-angle-ftir_multifile.tab"))])



import unittest

import numpy as np
import sklearn.model_selection as ms

import Orange
from Orange.classification import RandomForestLearner
from Orange.data import ContinuousVariable
from Orange.evaluation.testing import TestOnTestData
from Orange.evaluation.scoring import AUC
from Orange.data.table import DomainTransformationError

from orangecontrib.spectroscopy.preprocess import Interpolate, \
    Cut, SavitzkyGolayFiltering
from orangecontrib.spectroscopy.data import getx
from orangecontrib.spectroscopy.tests.util import smaller_data


learner = RandomForestLearner(random_state=42)

COLLAGEN = Orange.data.Table("collagen")
SMALL_COLLAGEN = smaller_data(COLLAGEN, 2, 2)


def separate_learn_test(data):
    sf = ms.ShuffleSplit(n_splits=1, test_size=0.2, random_state=np.random.RandomState(0))
    (traini, testi), = sf.split(y=data.Y, X=data.X)
    return data[traini], data[testi]


def slightly_change_wavenumbers(data, change):
    natts = [ContinuousVariable(float(a.name) + change) for a in data.domain.attributes]
    ndomain = Orange.data.Domain(natts, data.domain.class_vars,
                                 metas=data.domain.metas)
    ndata = data.transform(ndomain)
    with ndata.unlocked():
        ndata.X = data.X
    return ndata


def odd_attr(data):
    natts = [a for i, a in enumerate(data.domain.attributes) if i%2 == 0]
    ndomain = Orange.data.Domain(natts, data.domain.class_vars,
                                 metas=data.domain.metas)
    return data.transform(ndomain)


class TestConversion(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.collagen = SMALL_COLLAGEN

    def test_predict_same_domain(self):
        train, test = separate_learn_test(self.collagen)
        auc = AUC(TestOnTestData()(train, test, [learner]))
        self.assertGreater(auc, 0.9) # easy dataset

    def test_predict_different_domain(self):
        train, test = separate_learn_test(self.collagen)
        test = Interpolate(points=getx(test) - 1)(test) # other test domain
        with self.assertRaises(DomainTransformationError):
            learner(train)(test)

    def test_predict_different_domain_interpolation(self):
        train, test = separate_learn_test(self.collagen)
        aucorig = AUC(TestOnTestData()(train, test, [learner]))
        test = Interpolate(points=getx(test) - 1.)(test) # other test domain
        train = Interpolate(points=getx(train))(train)  # make train capable of interpolation
        aucshift = AUC(TestOnTestData()(train, test, [learner]))
        self.assertAlmostEqual(aucorig, aucshift, delta=0.01)  # shift can decrease AUC slightly
        test = Cut(1000, 1700)(test)
        auccut1 = AUC(TestOnTestData()(train, test, [learner]))
        test = Cut(1100, 1600)(test)
        auccut2 = AUC(TestOnTestData()(train, test, [learner]))
        test = Cut(1200, 1500)(test)
        auccut3 = AUC(TestOnTestData()(train, test, [learner]))
        # the more we cut the lower precision we get
        self.assertTrue(aucorig > auccut1 > auccut2 > auccut3)

    def test_predict_savgov_same_domain(self):
        data = SavitzkyGolayFiltering(window=9, polyorder=2, deriv=2)(self.collagen)
        train, test = separate_learn_test(data)
        auc = AUC(TestOnTestData()(train, test, [learner]))
        print(auc)
        self.assertGreater(auc, 0.85)

    def test_predict_savgol_another_interpolate(self):
        train, test = separate_learn_test(self.collagen)
        train = SavitzkyGolayFiltering(window=9, polyorder=2, deriv=2)(train)
        auc = AUC(TestOnTestData()(train, test, [learner]))
        train = Interpolate(points=getx(train))(train)
        aucai = AUC(TestOnTestData()(train, test, [learner]))
        self.assertAlmostEqual(auc, aucai, delta=0.02)



from unittest import TestCase

from Orange.data import Table

from orangecontrib.spectroscopy.preprocess.utils import table_eq_x


class TestEq(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.iris = Table("iris")
        cls.iris2 = Table("iris")
        cls.iris_changed = Table("iris")
        with cls.iris_changed.unlocked():
            cls.iris_changed[0][0] = 42

    def test_reference_eq_X_none(self):
        data = self.iris
        self.assertTrue(table_eq_x(None, None))
        self.assertFalse(table_eq_x(data, None))
        self.assertFalse(table_eq_x(None, data))

    def test_reference_eq_X_same(self):
        self.assertTrue(table_eq_x(self.iris, self.iris))
        self.assertTrue(table_eq_x(self.iris, self.iris2))
        self.assertFalse(table_eq_x(self.iris, self.iris_changed))



import numpy as np

import Orange.data
from Orange.widgets.widget import OWWidget, Msg, Input, Output
from Orange.widgets import gui, settings
from Orange.data.util import get_unique_names

from orangecontrib.spectroscopy.widgets.gui import lineEditIntOrNone

from AnyQt.QtWidgets import QWidget, QFormLayout


class OWReshape(OWWidget):
    # Widget's name as displayed in the canvas
    name = "Reshape Map"

    # Short widget description
    description = (
        "Builds or modifies the shape of the input dataset to create 2D maps "
        "from series data or change the dimensions of existing 2D datasets.")

    icon = "icons/reshape.svg"

    replaces = ["orangecontrib.infrared.widgets.owmapbuilder.OWMapBuilder",
                "orangecontrib.infrared.widgets.owreshape.OWReshape"]

    # Define inputs and outputs
    class Inputs:
        data = Input("Data", Orange.data.Table, default=True)

    class Outputs:
        map = Output("Map data", Orange.data.Table, default=True)

    autocommit = settings.Setting(True)

    want_main_area = False
    resizing_enabled = False

    settings_version = 2

    xpoints = settings.Setting(None)
    ypoints = settings.Setting(None)
    invert_x = settings.Setting(False)
    invert_y = settings.Setting(False)

    class Warning(OWWidget.Warning):
        wrong_div = Msg("Wrong divisor for {} curves.")
        nodata = Msg("No useful data on input!")

    def __init__(self):
        super().__init__()

        box = gui.widgetBox(self.controlArea, "Map grid")

        form = QWidget()
        formlayout = QFormLayout()
        form.setLayout(formlayout)
        box.layout().addWidget(form)

        self.le1 = lineEditIntOrNone(box, self, "xpoints", callback=self.le1_changed)
        formlayout.addRow("X dimension", self.le1)
        self.le3 = lineEditIntOrNone(box, self, "ypoints", callback=self.le3_changed)
        formlayout.addRow("Y dimension", self.le3)

        gui.checkBox(box, self, "invert_x", "Invert X axis",
                     callback=lambda: self.commit.deferred())
        gui.checkBox(box, self, "invert_y", "Invert Y axis",
                     callback=lambda: self.commit.deferred())

        gui.auto_commit(self.controlArea, self, "autocommit", "Send Data")

        self.data = None
        self.set_data(self.data)  # show warning

    @Inputs.data
    def set_data(self, dataset):
        self.Warning.wrong_div.clear()
        if dataset is not None:
            self.Warning.nodata.clear()
            self.data = dataset
        else:
            self.Warning.nodata()
        self.commit.now()

    # maybe doable with one callback...
    def le1_changed(self): # X dimension
        if self.data is not None and self.xpoints:
            self.Warning.wrong_div.clear()
            ytemp = len(self.data.X)//self.xpoints
            if len(self.data.X) % self.xpoints == 0:
                self.ypoints = ytemp
                self.commit.deferred()
            else:
                self.Warning.wrong_div(len(self.data.X))

    def le3_changed(self): # Y dimension
        if self.data is not None and self.ypoints:
            self.Warning.wrong_div.clear()
            xtemp = len(self.data.X)//self.ypoints
            if len(self.data.X) % self.ypoints == 0:
                self.xpoints = xtemp
                self.commit.deferred()
            else:
                self.Warning.wrong_div(len(self.data.X))
                # it would be nice to turn the bkg red at this point

    @gui.deferred
    def commit(self):
        map_data = None
        if self.data and self.xpoints is not None and self.ypoints is not None \
                and self.xpoints * self.ypoints == len(self.data):
            used_names = [var.name for var in self.data.domain.variables + self.data.domain.metas]
            xmeta = Orange.data.ContinuousVariable.make(get_unique_names(used_names, "X"))
            ymeta = Orange.data.ContinuousVariable.make(get_unique_names(used_names, "Y"))
            # add new variables for X and Y dimension ot the data domain
            metas = self.data.domain.metas + (xmeta, ymeta)
            domain = Orange.data.Domain(self.data.domain.attributes, self.data.domain.class_vars, metas)
            map_data = self.data.transform(domain)
            xpoints = np.arange(self.xpoints)
            ypoints = np.arange(self.ypoints)
            if self.invert_x:
                xpoints = np.flip(xpoints)
            if self.invert_y:
                ypoints = np.flip(ypoints)
            with map_data.unlocked(map_data.metas):
                map_data[:, xmeta] = np.tile(xpoints, len(self.data)//self.xpoints).reshape(-1, 1)
                map_data[:, ymeta] = np.repeat(ypoints, len(self.data)//self.ypoints).reshape(-1, 1)
        self.Outputs.map.send(map_data)

    def send_report(self):
        if self.xpoints and self.ypoints is not None:
            self.report_items((
                ("Number of points in the X direction", int(self.xpoints)),
                ("Number of points in the Y direction", int(self.ypoints))
            ))
        else:
            return


if __name__=="__main__":
    from Orange.widgets.utils.widgetpreview import WidgetPreview
    WidgetPreview(OWReshape).run(Orange.data.Table("collagen"))



import time
from typing import Optional

import numpy as np

from AnyQt.QtCore import Qt, QBuffer

from Orange.data import Table, ContinuousVariable, Domain
from Orange.data.util import get_unique_names
from Orange.widgets.settings import DomainContextHandler
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.widget import OWWidget, Input, Output, Msg
from Orange.widgets import gui, settings
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin

from orangewidget.settings import SettingProvider, ContextSetting

import io
from orangecontrib.spectroscopy.io.util import ConstantBytesVisibleImage
from PIL import Image
from orangecontrib.spectroscopy.utils import get_ndim_hyperspec, InvalidAxisException

from orangecontrib.spectroscopy.widgets.owhyper import BasicImagePlot


# Copied from orangecontrib.snom.widgets.owpreprocessimage
class AImagePlot(BasicImagePlot):
    attr_x = None  # not settings, set from the parent class
    attr_y = None

    def __init__(self, parent):
        super().__init__(parent)
        self.axes_settings_box.hide()
        self.rgb_settings_box.hide()
        self.image_updated.connect(self.drawing_done)

    def update_view(self):
        self.drawn = False
        super().update_view()

    def drawing_done(self):
        self.drawn = True

    def add_selection_actions(self, _):
        pass

    def clear_markings(self):
        pass


class InterruptException(Exception):
    pass


class OWOverlay(OWWidget, ConcurrentWidgetMixin):
    name = "Add Image Overlay"
    description = "Add an image that can be displayed in Hyper Spectra to the dataset."
    icon = "icons/overlay.svg"

    settings_version = 2
    value_type = 1

    settingsHandler = DomainContextHandler()
    imageplot = SettingProvider(AImagePlot)

    autocommit = settings.Setting(True)

    attr_value = ContextSetting(None)
    attr_x = ContextSetting(None, exclude_attributes=True)
    attr_y = ContextSetting(None, exclude_attributes=True)

    # Define inputs and outputs
    class Inputs:
        maindata = Input("Data", Table)
        data = Input("Overlay Data", Table)

    class Outputs:
        outdata = Output("Decorated Data", Table, default=True)

    class Error(OWWidget.Error):
        invalid_axis = Msg("Invalid axis: {}")
        invalid_block = Msg("Bin block size not compatible with dataset: {}")
        image_too_big = Msg("Image for chosen features is too big ({} x {}).")

    class Warning(OWWidget.Warning):
        nan_in_image = Msg("Unknown values within images: {} unknowns")
        threshold_error = Msg("Low slider should be less than High")

    class Information(OWWidget.Information):
        not_shown = Msg("Undefined positions: {} data point(s) are not shown.")

    def image_values(self):
        attr_value = self.attr_value.name if self.attr_value else None
        return lambda data, attr=attr_value: data.transform(Domain([data.domain[attr]]))

    def image_values_fixed_levels(self):
        return None

    def __init__(self):
        OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)

        self.maindata = None
        self.data = None

        # Imageplot for mainarea
        self.imageplot = AImagePlot(self)
        self.mainArea.layout().addWidget(self.imageplot)

        # Control area box
        self.imagebox = gui.widgetBox(self.controlArea, "Image")

        self.xy_model = DomainModel(
            DomainModel.METAS | DomainModel.CLASSES, valid_types=ContinuousVariable
        )

        self.feature_value_model = DomainModel(
            order=(
                DomainModel.ATTRIBUTES,
                DomainModel.Separator,
                DomainModel.CLASSES,
                DomainModel.Separator,
                DomainModel.METAS,
            ),
            valid_types=ContinuousVariable,
        )

        self.contextAboutToBeOpened.connect(lambda x: self._init_interface_data(x[0]))

        common_options = dict(
            labelWidth=50, orientation=Qt.Horizontal, sendSelectedValue=True
        )

        box = gui.vBox(self.imagebox)

        self.feature_value = gui.comboBox(
            box,
            self,
            "attr_value",
            label="Feature:",
            contentsLength=12,
            searchable=True,
            callback=self._update_feature_value,
            model=self.feature_value_model,
            **common_options,
        )

        gui.comboBox(
            box,
            self,
            f"attr_x",
            label=f"Axis X:",
            callback=self._attr_changed,
            model=self.xy_model,
            **common_options,
        )
        gui.comboBox(
            box,
            self,
            f"attr_y",
            label=f"Axis Y:",
            callback=self._attr_changed,
            model=self.xy_model,
            **common_options,
        )

        colorsbox = gui.widgetBox(self.controlArea, "Image colors")
        colorsbox.layout().addWidget(self.imageplot.color_settings_box)

        gui.rubber(self.controlArea)
        gui.auto_commit(self.controlArea, self, "autocommit", "Send Data")

        self.imageplot.color_cb.activated.connect(self.commit.deferred)
        self.imageplot._threshold_low_slider.sliderReleased.connect(
            self.commit.deferred
        )
        self.imageplot._threshold_high_slider.sliderReleased.connect(
            self.commit.deferred
        )
        self.imageplot._level_high_le.editingFinished.connect(self.commit.deferred)
        self.imageplot._level_low_le.editingFinished.connect(self.commit.deferred)
        self._init_attr_values(self.data)

    def _attr_changed(self):
        self._update_attrs()
        self.commit.deferred()

    def _init_attr_values(self, data):
        domain = data.domain if data is not None else None
        self.feature_value_model.set_domain(domain)
        self.attr_value = (
            self.feature_value_model[0] if self.feature_value_model else None
        )
        self.xy_model.set_domain(domain)
        self.attr_x = self.xy_model[0] if self.xy_model else None
        self.attr_y = self.xy_model[1] if len(self.xy_model) >= 2 else self.attr_x

    def _init_interface_data(self, data):
        self._init_attr_values(data)
        self.imageplot.init_interface_data(data)

    def _update_feature_value(self):
        self.redraw_data()

    def _update_attrs(self):
        self.imageplot.attr_x = self.attr_x
        self.imageplot.attr_y = self.attr_y
        self.redraw_data()

    def redraw_data(self):
        self.imageplot.update_view()
        self.commit.deferred()

    @staticmethod
    def with_overlay(imageplot, data, maindata, state):

        if data is None or maindata is None:
            return None

        def progress_interrupt():
            if state.is_interruption_requested():
                raise InterruptException

        # wait for the image to appear
        while True:
            time.sleep(0.010)
            progress_interrupt()
            if imageplot.drawn is True:
                break

        # the following could raise an InvalidAxisException
        hypercube, ls = get_ndim_hyperspec(
            data, (imageplot.attr_x, imageplot.attr_y)
        )

        # if drawing failed for some reason
        if imageplot.img.image is None:
            return None

        # Copy main data to avoid changing it in place
        newmaindata = maindata.copy()

        width = np.abs(ls[0][1] - ls[0][0])
        height = np.abs(ls[1][1] - ls[1][0])
        xres = width / hypercube.shape[:2][0]
        yres = height / hypercube.shape[:2][1]
        posx = ls[0][0]
        posy = ls[1][0]

        # Extract the image from imageplot so the coloring is done
        imageplot.img.render()  # ensures qimage is produced even when not displayed
        im = imageplot.img.qimage
        buffer = QBuffer()
        buffer.open(QBuffer.ReadWrite)
        im.save(buffer, "PNG")
        pil_im = Image.open(io.BytesIO(buffer.data()))
        pil_im = pil_im.transpose(Image.FLIP_TOP_BOTTOM)
        img_bytes = io.BytesIO()
        pil_im.save(img_bytes, format="PNG")
        # Update name
        allnames = []
        if "visible_images" in newmaindata.attributes:
            allnames = [
                im.name for im in newmaindata.attributes["visible_images"]
            ]
        basename = "Overlay Image"
        name = get_unique_names(names=allnames, proposed=basename)
        # Need to modify the position and the scale since visual imageplot
        # place the corner of the pixel to the given position
        vimage = ConstantBytesVisibleImage(
            name=name,
            pos_x=posx - xres / 2,
            pos_y=posy - yres / 2,
            size_x=width + xres,
            size_y=height + yres,
            image_bytes=img_bytes,
        )

        # # Assign it to the datatable attributes
        if newmaindata and vimage is not None:
            if "visible_images" in newmaindata.attributes:
                newmaindata.attributes["visible_images"].append(vimage)
            else:
                newmaindata.attributes["visible_images"] = [vimage]

        return newmaindata

    @Inputs.maindata
    def set_data(self, data):
        self.maindata = data

    @Inputs.data
    def set_overlaydata(self, data):
        self.closeContext()
        self.openContext(data)
        self.data = data

        self.Warning.nan_in_image.clear()
        self.Error.invalid_axis.clear()
        self.Error.invalid_block.clear()

        self.imageplot.set_data(data)
        self.imageplot.update_view()

    def handleNewSignals(self):
        self.commit.now()

    @gui.deferred
    def commit(self):
        self.cancel()
        self.Warning.nan_in_image.clear()
        self.Error.invalid_axis.clear()
        self.Error.invalid_block.clear()
        self.start(OWOverlay.with_overlay, self.imageplot, self.data, self.maindata)

    def on_partial_result(self, _):
        pass

    def on_done(self, result: Optional[Table]):
        self.Outputs.outdata.send(result)

    def on_exception(self, ex):
        if isinstance(ex, InvalidAxisException):
            self.Error.invalid_axis(str(ex))
        self.Outputs.outdata.send(None)

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()


if __name__ == "__main__":  # pragma: no cover
    from Orange.widgets.utils.widgetpreview import WidgetPreview

    WidgetPreview(OWOverlay).run(
        set_data=Table("agilent/5_mosaic_agg1024.dmt"),
        set_overlaydata=Table("agilent/5_mosaic_agg1024.dmt"),
    )



from typing import List

import spectral.io
from Orange.data import FileFormat

from orangecontrib.spectroscopy.io.agilent import AgilentImageReader
from orangecontrib.spectroscopy.io.ascii import AsciiColReader
from orangecontrib.spectroscopy.io.cls import HDF5Reader_SGM
from orangecontrib.spectroscopy.io.envi import EnviMapReader
from orangecontrib.spectroscopy.io.maxiv import HDRReader_STXM
from orangecontrib.spectroscopy.io.soleil import HDF5Reader_ROCK


class DatMetaReader(FileFormat):
    """ Meta-reader to handle agilentImageReader and AsciiColReader name clash
    over .dat extension. """
    EXTENSIONS = ('.dat',)
    DESCRIPTION = 'Spectra ASCII or Agilent Single Tile Image'
    PRIORITY = min(AsciiColReader.PRIORITY, AgilentImageReader.PRIORITY) - 1

    def read(self):
        try:
            # agilentImage requires the .bsp file to be present as well
            return AgilentImageReader(filename=self.filename).read()
        except OSError:
            return AsciiColReader(filename=self.filename).read()


class HDF5MetaReader(FileFormat):
    """ Meta-reader to handle HDF5 (currently just .h5) extension(s)"""
    EXTENSIONS = ('.h5',)
    DESCRIPTION = "HDF5 files (ROCK/SGM)"
    PRIORITY = min(HDF5Reader_ROCK.PRIORITY, HDF5Reader_SGM.PRIORITY) - 1

    @property
    def sheets(self) -> List:
        sheets = []
        for reader in [HDF5Reader_SGM, HDF5Reader_ROCK]:
            try:
                sheets.extend(reader(self.filename).sheets)
            except:
                pass
        return sheets

    def read(self):
        try:
            reader = HDF5Reader_SGM(self.filename)
            reader.sheet = self.sheet
            return reader.read()
        except (ValueError, IndexError):
            return HDF5Reader_ROCK(filename=self.filename).read()

class HDRMetaReader(FileFormat):
    """ Meta-reader to handle EnviMapReader and HDRReader_STXM name clash
    over .hdr extension. """
    EXTENSIONS = ('.hdr',)
    DESCRIPTION = 'Envi hdr or STXM hdr+xim files'
    PRIORITY = min(EnviMapReader.PRIORITY, HDRReader_STXM.PRIORITY) - 1

    def read(self):
        try:
            return EnviMapReader(filename=self.filename).read()
        except spectral.io.envi.FileNotAnEnviHeader:
            return HDRReader_STXM(filename=self.filename).read()


from orangecontrib.spectroscopy.tests.test_owpreprocess import PreprocessorEditorTest
from orangecontrib.spectroscopy.widgets.owpreprocess import OWPreprocess
from orangecontrib.spectroscopy.tests.test_preprocess import SMALL_COLLAGEN
from orangecontrib.spectroscopy.widgets.preprocessors.misc import CutEditor
from orangecontrib.spectroscopy.preprocess import Cut


class TestCutEditor(PreprocessorEditorTest):

    def setUp(self):
        self.widget = self.create_widget(OWPreprocess)
        self.editor = self.add_editor(CutEditor, self.widget)
        self.data = SMALL_COLLAGEN
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_for_preview()  # ensure initialization with preview data

    def test_no_interaction(self):
        p = self.commit_get_preprocessor()
        self.assertIsInstance(p, Cut)
        self.assertEqual(p.lowlim, 995.902)
        self.assertEqual(p.highlim, 1711.779)
        self.assertEqual(p.inverse, False)

    def test_basic(self):
        self.editor.lowlim = 5
        self.editor.highlim = 10
        self.editor.inverse = True
        self.editor.edited.emit()
        p = self.commit_get_preprocessor()
        self.assertEqual(p.lowlim, 5)
        self.assertEqual(p.highlim, 10)
        self.assertEqual(p.inverse, True)



import struct
import numpy as np

# This file is based on the development chain indicated below:

# The below file readers are based on file readers written by Specio
# (https://github.com/paris-saclay-cds/specio) with the following
# license:
# BSD 3-Clause License

# Copyright (c) 2017, Guillaume Lemaitre
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# This code is partially based on software developed in the Diamond synchrotron.

class BlockReader:
    def __init__(self, data, start=0):
        self.data = data
        self.start = start

    def peek(self, step, format=None, expect_tuple=False):
        block = self.data[self.start: self.start + step]
        return BlockReader.format(block,
                                  format,
                                  expect_tuple=expect_tuple)

    def read(self, step, format=None, expect_tuple=False):
        block = self.peek(step,
                          format=format,
                          expect_tuple=expect_tuple)
        self.step(step)
        return block

    def step(self, step):
        self.start += step

    def atEnd(self, step=0):
        return self.start >= self.size - step

    @property
    def size(self):
        return len(self.data)

    @staticmethod
    def readData(data, start, step, format=None, expect_tuple=False):
        block = data[start: start + step]
        return BlockReader.format(block,
                                  format,
                                  expect_tuple=expect_tuple)

    @staticmethod
    def format(data, format, expect_tuple=False):
        DECODE_FORMATS = [
            "utf-8",
        ]

        if format == None:
            return data

        if format in DECODE_FORMATS:
            assert (not expect_tuple)

            return BlockReader._decode(data, format)

        unpacked = BlockReader._unpack(data, format)

        if expect_tuple:
            return unpacked

        return BlockReader._asSingular(unpacked)

    @staticmethod
    def _decode(data, format):
        return data.decode(format)

    @staticmethod
    def _unpack(data, format):
        return struct.unpack(format, data)

    @staticmethod
    def _asSingular(unpacked):
        assert (len(unpacked) == 1)

        return unpacked[0]


class PerkinElmer:
    @staticmethod
    def varIdDecode(data, var_id, format, step, expect_tuple=False):
        _var_id = BlockReader.readData(data, 0, 2, "<H", False)

        if var_id == _var_id:
            return BlockReader.readData(data, 2, step, format=format,
                                        expect_tuple=expect_tuple)

        raise struct.error(f"Expected {var_id}, but got {_var_id}.")

    @staticmethod
    def createMeta(values, attrs, meta=dict()):
        for attr_name, attr_index in attrs:
            try:
                meta[attr_name] = values[attr_index]
            except IndexError:
                meta[attr_name] = None

        return meta

    @staticmethod
    def decode5100(data):
        name_size = BlockReader.readData(data,
                                         0,
                                         2,
                                         format="<h",
                                         expect_tuple=False)

        name = BlockReader.readData(data, 2, name_size, format="utf-8")

        header_format = '<ddddddddddiiihBhBhBhB'
        header_size = 104

        values = BlockReader.readData(data, name_size + 2,
                                      header_size,
                                      format=header_format,
                                      expect_tuple=True)

        meta = PerkinElmer.createMeta(
            values,
            [
                ("x_delta", 0),
                ("y_delta", 1),
                ("z_delta", 2),
                ("z_start", 3),
                ("z_end", 4),
                ("z_4d_start", 5),
                ("z_4d_end", 6),
                ("x_init", 7),
                ("y_init", 8),
                ("z_init", 9),
                ("n_x", 10),
                ("n_y", 11),
                ("n_z", 12),
                ("text1", 14),
                ("text2", 16),
                ("resolution", 17),
                ("text3", 18),
                ("transmission", 19),
                ("text4", 20),
            ],
            meta={"name": name, }
        )

        return meta

    @staticmethod
    def decode5104(data):
        values = []

        block_reader = BlockReader(data)

        while not block_reader.atEnd(2):
            tag = block_reader.read(2)

            if tag == b'#u':
                value_size = block_reader.read(2,
                                               format="<h",
                                               expect_tuple=False)

                values.append(block_reader.read(value_size,
                                                format="utf-8"))

                block_reader.step(6)

            elif tag == b'$u':
                values.append(block_reader.read(2,
                                                format="<h",
                                                expect_tuple=False))

                block_reader.step(6)

            elif tag == b',u':
                values.append(block_reader.read(2,
                                                format="<h",
                                                expect_tuple=False))

            else:
                block_reader.step(-1)

        meta = PerkinElmer.createMeta(
            values,
            [
                ("analyst", 0),
                ("date", 2),
                ("image_name", 4),
                ("instrument_model", 5),
                ("instrument_serial_number", 6),
                ("instrument_software_version", 7),
                ("accumulations", 9),
                ("detector", 11),
                ("source", 12),
                ("beam_splitter", 13),
                ("apodization", 15),
                ("spectrum_type", 16),
                ("beam_type", 17),
                ("phase_correction", 20),
                ("ir_accessory", 26),
                ("igram_type", 28),
                ("scan_direction", 29),
                ("background_scans", 32),
                ("ir_laser_wave_number_unit", 67),
            ],
        )

        return meta

    @staticmethod
    def decode5105(data):
        return np.frombuffer(data, dtype=np.float32)

    @staticmethod
    def decode25739(data):
        var_size = PerkinElmer.varIdDecode(data,
                                           29987,
                                           "<H",
                                           2,
                                           expect_tuple=False)

        file_path = BlockReader.readData(data,
                                         4,
                                         var_size,
                                         format="utf-8")

        return {'file_path': file_path}

    @staticmethod
    def decode35698(data):
        min_w, max_w = PerkinElmer.varIdDecode(data,
                                               29981,
                                               "<dd",
                                               16,
                                               expect_tuple=True)

        return {'min_wavelength': min_w, 'max_wavelength': max_w}

    @staticmethod
    def decode35699(data):
        min_abs, max_abs = PerkinElmer.varIdDecode(data,
                                                   29981,
                                                   "<dd",
                                                   16,
                                                   expect_tuple=True)

        return {'min_absolute': min_abs, 'max_absolute': max_abs}

    @staticmethod
    def decode35700(data):
        wavelength_step = PerkinElmer.varIdDecode(data,
                                                  29979,
                                                  "<d",
                                                  8,
                                                  expect_tuple=False)

        return {'wavelength_step': wavelength_step}

    @staticmethod
    def decode35701(data):
        n_points = PerkinElmer.varIdDecode(data,
                                           29995,
                                           "<I",
                                           4,
                                           expect_tuple=False)

        return {'n_points': n_points}

    @staticmethod
    def decode35708(data):
        var_size = PerkinElmer.varIdDecode(data,
                                           29974,
                                           "<I",
                                           4,
                                           expect_tuple=False)

        return np.frombuffer(
            BlockReader.readData(data, 6, var_size, format=None),
            dtype=np.float64
        )
